# -*- coding: utf-8 -*-
"""click_loguru -- Setup loguru logging with stderr and file with click."""

# standard library imports
import functools
import sys
from datetime import datetime
from pathlib import Path
from time import process_time

# third-party imports
import attr
from click import get_current_context as cur_ctx
from click import option
from loguru import logger
from memory_profiler import memory_usage

# global constants
__version__ = "1.3.0"
__all__ = ["ClickLoguru"]
DEFAULT_STDERR_LOG_LEVEL = "INFO"
DEFAULT_FILE_LOG_LEVEL = "DEBUG"
NO_LEVEL_BELOW = 30  # Don't print level for messages below this level
SKIP_FIELDS = -7


class ClickLoguru:
    """Creates decorators for use with click to control loguru logging ."""

    @attr.s(auto_attribs=True)
    class LogState:
        """Click context object for verbosity, quiet, and logfile info."""

        verbose: bool = False
        quiet: bool = False
        logfile: bool = True
        profile_mem: bool = True
        logfile_path: Path = None
        logfile_handler_id: int = None
        subcommand: str = None
        user_options: attr.s = {}
        max_mem: int = 0

    def __init__(
        self,
        name,
        version,
        retention=None,
        stderr_format_func=None,
        log_dir_parent=None,
        file_log_level=DEFAULT_FILE_LOG_LEVEL,
        stderr_log_level=DEFAULT_STDERR_LOG_LEVEL,
        timer_log_level="debug",
    ):
        """Initialize logging setup info."""
        self._name = name
        self._version = version
        self._retention = retention
        self._log_dir_parent = log_dir_parent
        self._file_log_level = file_log_level
        self._stderr_log_level = stderr_log_level
        self.timer_log_level = timer_log_level.upper()
        self.start_times = {
            "Total": {"wall": datetime.now(), "process": process_time()}
        }
        self.phase = None
        if stderr_format_func is None:

            def format_func(msgdict):
                """Do level-sensitive formatting."""
                if msgdict["level"].no < NO_LEVEL_BELOW:
                    return "<level>{message}</level>\n"
                return "<level>{level}</level>: <level>{message}</level>\n"

            self.stderr_format_func = format_func
        else:
            self.stderr_format_func = stderr_format_func

    def _verbose_option(self, user_func):
        """Define verbose option."""

        def callback(ctx, unused_param, value):
            """Set verbose state."""
            state = ctx.ensure_object(self.LogState)
            state.verbose = value
            return value

        return option(
            "-v",
            "--verbose",
            is_flag=True,
            show_default=True,
            default=False,
            help="Log debugging info to stderr.",
            callback=callback,
        )(user_func)

    def _quiet_option(self, user_func):
        """Define quiet option."""

        def callback(ctx, unused_param, value):
            """Set quiet state."""
            state = ctx.ensure_object(self.LogState)
            state.quiet = value
            return value

        return option(
            "-q",
            "--quiet",
            is_flag=True,
            show_default=True,
            default=False,
            help="Suppress info to stderr.",
            callback=callback,
        )(user_func)

    def _logfile_option(self, user_func):
        """Define logfile option."""

        def callback(ctx, unused_param, value):
            """Set logfile state."""
            state = ctx.ensure_object(self.LogState)
            state.logfile = value
            return value

        return option(
            "--logfile/--no-logfile",
            is_flag=True,
            show_default=True,
            default=True,
            help="Log to file.",
            callback=callback,
        )(user_func)

    def _profile_mem_option(self, user_func):
        """Define logfile option."""

        def callback(ctx, unused_param, value):
            """Set logfile state."""
            state = ctx.ensure_object(self.LogState)
            state.profile_mem = value
            return value

        return option(
            "--profile_mem",
            is_flag=True,
            show_default=True,
            default=False,
            help="Profile peak memory use.",
            callback=callback,
        )(user_func)

    def logging_options(self, user_func):
        """Set all logging options."""
        user_func = self._verbose_option(user_func)
        user_func = self._quiet_option(user_func)
        user_func = self._logfile_option(user_func)
        user_func = self._profile_mem_option(user_func)
        return user_func

    def init_logger(self, log_dir_parent=None, logfile=True):
        """Log to stderr and to logfile at different levels."""

        def decorator(user_func):
            @functools.wraps(user_func)
            def wrapper(*args, **kwargs):
                state = cur_ctx().find_object(self.LogState)
                # get the verbose/quiet levels from context
                if state.verbose:
                    log_level = "DEBUG"
                elif state.quiet:
                    log_level = "ERROR"
                else:
                    log_level = self._stderr_log_level
                logger.remove()  # remove existing default logger
                logger.add(
                    sys.stderr, level=log_level, format=self.stderr_format_func
                )
                if logfile and state.logfile:  # start a log file
                    # If a subcommand was used, log to a file in the
                    # logs/ subdirectory with the subcommand in the file name.
                    if log_dir_parent is not None:
                        self._log_dir_parent = log_dir_parent
                    if self._log_dir_parent is None:
                        log_dir_path = Path(".") / "logs"
                    else:
                        log_dir_path = Path(self._log_dir_parent)
                    subcommand = cur_ctx().invoked_subcommand
                    if subcommand is None:
                        subcommand = state.subcommand
                    if subcommand is not None:
                        logfile_prefix = f"{self._name}-{subcommand}"
                    else:
                        logfile_prefix = f"{self._name}"
                    if log_dir_path.exists():
                        log_numbers = [
                            f.name[len(logfile_prefix) + 1 : -4]
                            for f in log_dir_path.glob(
                                logfile_prefix + "_*.log"
                            )
                        ]
                        log_number_ints = sorted(
                            [int(n) for n in log_numbers if n.isnumeric()]
                        )
                        if len(log_number_ints) > 0:
                            log_number = log_number_ints[-1] + 1
                            if (
                                self._retention is not None
                                and len(log_number_ints) > self._retention
                            ):
                                for remove in log_number_ints[
                                    : len(log_number_ints) - self._retention
                                ]:
                                    (
                                        log_dir_path
                                        / f"{logfile_prefix}_{remove}.log"
                                    ).unlink()
                        else:
                            log_number = 0
                    else:
                        log_number = 0
                    if self._retention == 0:
                        state.logfile_path = (
                            log_dir_path / f"{logfile_prefix}.log"
                        )
                    else:
                        state.logfile_path = (
                            log_dir_path / f"{logfile_prefix}_{log_number}.log"
                        )
                    state.logfile_handler_id = logger.add(
                        str(state.logfile_path), level=self._file_log_level
                    )
                logger.debug(f'Command line: "{" ".join(sys.argv)}"')
                logger.debug(f"{self._name} version {self._version}")
                logger.debug(
                    f"Run started at {str(self.start_times['Total']['wall'])[:SKIP_FIELDS]}"
                )
                return user_func(*args, **kwargs)

            return wrapper

        return decorator

    def log_elapsed_time(self, level="debug"):
        """Log the elapsed time for (sub)command."""

        def decorator(user_func):
            @functools.wraps(user_func)
            def wrapper(*args, **kwargs):
                returnobj = user_func(*args, **kwargs)
                logger.log(
                    level.upper(), self._format_time("Total"),
                )
                return returnobj

            return wrapper

        return decorator

    def log_peak_memory_use(self, level="debug"):
        """Log the peak memory use for (sub)command."""

        def decorator(user_func):
            @functools.wraps(user_func)
            def wrapper(*args, **kwargs):
                state = cur_ctx().find_object(self.LogState)
                if state.profile_mem:
                    max_mem, returnobj = memory_usage(
                        (user_func, args, kwargs),
                        retval=True,
                        include_children=True,
                        max_usage=True,
                        multiprocess=True,
                        max_iterations=1,
                    )
                    state.max_mem = int(max_mem)
                    logger.log(
                        level.upper(),
                        f"Peak total memory use = {state.max_mem} MB.",
                    )
                else:
                    returnobj = user_func(*args, **kwargs)
                return returnobj

            return wrapper

        return decorator

    def stash_subcommand(self):
        """Save the subcommand to the context object."""

        def decorator(user_func):
            @functools.wraps(user_func)
            def wrapper(*args, **kwargs):
                state = cur_ctx().find_object(self.LogState)
                state.subcommand = cur_ctx().invoked_subcommand
                return user_func(*args, **kwargs)

            return wrapper

        return decorator

    def get_global_options(self):
        """Return dictionary of global options."""
        return cur_ctx().find_object(self.LogState)

    def get_user_global_options(self):
        """Return dict of global user options."""
        return cur_ctx().find_object(self.LogState).user_options

    def user_global_options_callback(self, ctx, param, value):
        """Put user global options in user dict."""
        state = ctx.ensure_object(self.LogState)
        state.user_options[param.name] = value
        return value

    def elapsed_time(self, phase):
        """Log the elapsed time of a phase."""
        old_phase = self.phase
        if phase is None:
            self.phase = None
        else:
            self.phase = phase.capitalize()
            self.start_times[self.phase] = {
                "wall": datetime.now(),
                "process": process_time(),
            }
        if old_phase is None:
            return
        logger.log(
            self.timer_log_level, self._format_time(old_phase),
        )

    def _format_time(self, phase_name):
        """Return a formatted elapsed time string."""
        now = datetime.now()
        wall = str(datetime.now() - self.start_times[phase_name]["wall"])[
            :SKIP_FIELDS
        ]
        cpu = process_time() - self.start_times[phase_name]["process"]
        return f"{phase_name} elapsed time is {wall}, {cpu:.1f} s CPU"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['click_loguru']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3,<21.0',
 'click>=7.1.2,<8.0.0',
 'loguru>=0.5.0,<0.6.0',
 'memory_profiler>=0.57.0,<0.58.0']

setup_kwargs = {
    'name': 'click-loguru',
    'version': '1.3.0',
    'description': 'Logging to stderr and file for click applications.',
    'long_description': 'click_loguru\n============\n``click_loguru`` initializes `click <https://click.palletsprojects.com/>`_ CLI-based\nprograms for logging to stderr and (optionally) a log file via the\n`loguru <https://github.com/Delgan/loguru/>`_ logger.  It can optionally log run time,\nCPU use, and peak memory use of user functions.\n\nLog file names will include the name of your program and (if your application uses\nsubcommands via ``@click.group()``), the name of the subcommand. Log files are \n(optionally) numbered, with a retention policy specified.  Log files can be\nenabled or disabled per-subcommand and written to a subdirectory that your\napplication specifies.  \n\nGlobal CLI options control verbose/quiet levels and log file creation.\nThe values of these global options are accessible, along with the path to the\nlog file, from your application.\n\nInstantiation\n-------------\n``click_loguru`` objects are instantiated from the ``ClickLoguru`` class as::\n\n      click_loguru = ClickLoguru(name,\n                                 version,\n                                 retention=4,\n                                 stderr_format_func=None,\n                                 log_dir_parent="./logs",\n                                 file_log_level="DEBUG",\n                                 stderr_log_level="INFO",\n                                 timer_log_level="debug",\n        )\n\nwhere:\n\n* **name** is the name of your application\n* **version** is the version string of your application\n* **retention** is the log file retention policy.  If set to a non-zero value, the\n  log files will be given by ``logs/NAME[-SUBCOMMAND]_n.log`` where ```NAME`` is the name\n  of your application, ``SUBCOMMAND`` is the group subcommand (if you are using\n  click groups), and ``n`` is an integer number.  The value of ``retention`` specifies\n  the number of log files to be kept.\n* **stderr_format_func** is the format function to be used for messages to stderr, as\n  defined by ``loguru``.  Default is very short, with ``INFO``-level messages having\n  no level name printed.\n* **log_dir_parent** sets the location of the log file directory.  This value may be\n  overridden per-command.\n* **file_log_level**  sets the level of logging to the log file.\n* **stderr_log_level** sets the level of logging to stderr.  This value may be overridden\n  by the ``--quiet`` or ``--verbose`` options.\n* **timer_log_level** is the level at which ``elapsed_time`` results will be logged.\n\n\nMethods\n-------\nThe ``ClickLoguru`` class defines the following methods:\n\n* **logging_options** is a decorator to be used for your application\'s CLI function.  This\n  decorator defines the global options that allows control of ``quiet``, ``verbose``,\n  and ``log file`` booleans.\n\n* **stash_subcommand** is a  decorator to be used for the CLI method for applications\n  which define subcommands.\n\n* **init_logger** is  a decorator which must be used for each subcommand.   It allows\n  override of the default ``log_dir_parent`` established at instantiation,\n  as well as turning off file logging for that command by setting ``log file`` to ``False``.\n\n* **log_elapsed_time** is a decorator which causes the elapsed wall-clock time and\n  CPU time in seconds for the (sub)command\n  to be emitted at the level specified by the ``level=`` argument (``debug`` by default).\n\n* **get_global_options** is a method that returns the context object associated with the\n  global options. The context object is printable.  The attributes of the context object are the booleans ``verbose``,\n  ``quiet``, and ``log file``, the string ``subcommand`` showing the subcommand that was invoked,\n  and ``log file_handler_id`` if your code wishes to manipulate the handler directly.\n\n* **user_global_options_callback** is a method to be used as\n  a callback when your code declares a global option.  Values\n  of these global options will be stored in a user global\n  options context dictionary.\n\n* **get_user_global_options** is a method to retrieve a\n  dictionary of values of user global options.\n\n* **elapsed_timer** is a method that accepts a single argument, ``phase``.\n  The next invocation of this method will produce a log entry at ``timer_log_level``\n  showing the elapsed wall clock and CPU time.  If ``phase`` is ``None``, \n  the next invocation will not produce a message.\n\n* **log_peak_memory_use** is a method that results in the peak memory usage for\n  the function and children of the function to be emitted at a level specified\n  by the ``level=`` keyword (``debug`` is default).  This functionality\n  is somewhat expensive in that it requires an additional thread, so the global\n  option ``--profile_mem`` must be enabled.\n\n\nSee the `simple test CLI application\n<https://github.com/legumeinfo/click_loguru/blob/master/tests/__init__.py>`_\nfor usage examples.\n\nPrerequisites\n-------------\nPython 3.6 or greater is required.\nThis package is tested under Linux using Python 3.8.  This package has\nno dependencies beyond``click`` and ``loguru`` themselves.\n\n\nProject Status\n--------------\n+-------------------+------------+\n| Latest Release    | |pypi|     |\n+-------------------+------------+\n| Activity          | |repo|     |\n+-------------------+------------+\n| Downloads         ||downloads| |\n+-------------------+------------+\n| Download Rate     | |dlrate|   |\n+-------------------+------------+\n| License           | |license|  |\n+-------------------+------------+\n| Travis Build      | |travis|   |\n+-------------------+------------+\n| Coverage          | |coverage| |\n+-------------------+------------+\n| Code Grade        | |codacy|   |\n+-------------------+------------+\n| Dependencies      | |depend|   |\n+-------------------+------------+\n| Issues            | |issues|   |\n+-------------------+------------+\n| Code Style        | |black|    |\n+-------------------+------------+\n| Pre-commit        | |precommit||\n+-------------------+------------+\n\n.. |pypi| image:: https://img.shields.io/pypi/v/click_loguru.svg\n    :target: https://pypi.python.org/pypi/click_loguru\n    :alt: Python package\n\n.. |repo| image:: https://img.shields.io/github/last-commit/legumeinfo/click_loguru\n    :target: https://github.com/legumeinfo/click_loguru\n    :alt: GitHub repository\n\n.. |license| image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg\n    :target: https://github.com/legumeinfo/click_loguru/blob/master/LICENSE.txt\n    :alt: License terms\n\n.. |travis| image:: https://img.shields.io/travis/legumeinfo/click_loguru.svg\n    :target:  https://travis-ci.org/legumeinfo/click_loguru\n    :alt: Travis CI\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/6ee5771afe014cffbb32a2f79cf17fff\n    :target: https://www.codacy.com/gh/legumeinfo/click_loguru?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=legumeinfo/click_loguru&amp;utm_campaign=Badge_Grade\n    :alt: Codacy.io grade\n\n.. |coverage| image:: https://codecov.io/gh/legumeinfo/click_loguru/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/legumeinfo/click_loguru\n    :alt: Codecov.io test coverage\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square\n    :target: https://github.com/psf/black\n    :alt: Black is the uncompromising Python code formatter\n\n.. |precommit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n    :target: https://github.com/pre-commit/pre-commit\n    :alt: pre-commit\n\n.. |issues| image:: https://img.shields.io/github/issues/legumeinfo/click_loguru.svg\n    :target:  https://github.com/legumeinfo/click_loguru/issues\n    :alt: Issues reported\n\n\n.. |depend| image:: https://api.dependabot.com/badges/status?host=github&repo=legumeinfo/click_loguru\n     :target: https://app.dependabot.com/accounts/legumeinfo/repos/236847525\n     :alt: dependabot dependencies\n\n\n.. |dlrate| image:: https://img.shields.io/pypi/dm/click_loguru\n    :target: https://pypistats.org/packages/click_loguru\n    :alt: Download stats\n\n.. |downloads| image:: https://pepy.tech/badge/click_loguru\n    :target: https://pepy.tech/project/click_loguru\n    :alt: Download stats\n',
    'author': 'Joel Berendzen',
    'author_email': 'joelb@ncgr.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/legumeinfo/click_loguru',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
